from methurator.plot_utils.plot_curve import plot_curve
import rich_click as click
from methurator.config_utils.config_formatter import ConfigFormatter
from methurator.config_utils.validation_utils import (
    validate_read_summary,
    validate_cpgs_summary,
)
from rich.console import Console
from rich.panel import Panel

console = Console()


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option(
    "--cpgs_file",
    "-c",
    type=click.Path(exists=True),
    required=True,
    help="File containing CpGs coverage information.",
)
@click.option(
    "--reads_file",
    "-r",
    type=click.Path(exists=True),
    required=True,
    help="File containing reads coverage information.",
)
@click.option(
    "--outdir",
    type=click.Path(),
    default="output",
    help="Default output directory.",
)
@click.option("-v", "--verbose", is_flag=True, help="Enable verbose logging.")
def plot(**kwargs):

    # Import and validate params
    configs = ConfigFormatter(**kwargs)
    validate_cpgs_summary(configs.cpgs_file)
    validate_read_summary(configs.reads_file)

    # Print I/O parameters
    params_text = ""
    params_text += (
        f"[purple]CpGs summary file:[/purple] [blue]{configs.cpgs_file}[/blue]\n"
    )
    params_text += (
        f"[purple]Reads summary file:[/purple] [blue]{configs.reads_file}[/blue]\n"
    )
    params_text += f"[purple]Output directory:[/purple] [blue]{configs.outdir}[/blue]"
    console.print(
        Panel(
            params_text,
            title="📌 [bold cyan]Input / Output Parameters[/bold cyan]",
            border_style="cyan",
            expand=False,
        )
    )

    # Fit the model and plot the saturation curve
    plot_curve(configs)



import os
import glob
import rich_click as click
from methurator.config_utils.verbose_utils import vprint
from methurator.config_utils.validation_utils import ensure_coordinated_sorted


def bam_to_list(configs):

    # If both bamdir and bam are provided, use the bamdir parameter
    if configs.bamdir and configs.bam:
        vprint(
            "[yellow]⚠️ Warning: both --bam and --bamdir were provided. Only --bamdir will be considered.[/yellow]",
            True,
        )
    csorted_bams = []

    # If bamdir is provided, import all bam files in the directory
    if configs.bamdir:
        raw_bam_files = import_bam_files(configs.bamdir)
        vprint(
            f"[bold]Found {len(raw_bam_files)} BAM files in directory '{configs.bamdir}'.[/bold]",
            configs.verbose,
        )
        for bam in raw_bam_files:
            # Check if bam file coordinate-sorted or sort it
            try:
                csorted_bam = ensure_coordinated_sorted(configs)
                csorted_bams.append(csorted_bam)
            except ValueError as e:
                raise click.UsageError(f"{e}")
    else:
        try:
            csorted_bam = ensure_coordinated_sorted(configs)
            csorted_bams.append(csorted_bam)
        except ValueError as e:
            raise click.UsageError(f"{e}")

    return csorted_bams


def import_bam_files(bam_dir):

    # Get all BAM files inside directory
    bam_files_all = sorted(glob.glob(os.path.join(bam_dir, "*.bam")))

    # Check if any BAM files were found
    if not bam_files_all:
        raise click.UsageError(f"No BAM files found in directory: {bam_dir}")

    # Remove duplicates (based on filename) and warn
    seen_filenames = set()
    bam_files = []
    for bam in bam_files_all:
        fname = os.path.basename(bam)
        if fname in seen_filenames:
            vprint(
                f"[yellow]⚠️ Warning: Duplicate BAM file found and discarded: {fname}[/yellow].",
                True,
            )
        else:
            seen_filenames.add(fname)
            bam_files.append(bam)

    # bam_files now contains only the BAM files
    return bam_files



import rich_click as click
from methurator.config_utils.verbose_utils import vprint
from methurator.config_utils.download_reference import get_reference
import os
import subprocess
import pysam
import pandas as pd


def mincoverage_checker(coverages):
    """
    Converts a comma-separated string into a list of integers.
    """
    values = coverages.split(",")
    list_coverages = []

    for x in values:
        x = x.strip()
        if not x.isdigit():
            raise click.UsageError(
                f"Invalid minimum coverage value: '{x}'. All minimum coverage must be integers."
            )
        if int(x) == 0:
            vprint(
                f"[yellow]⚠️ Warning: coverage values must be at least >=1, '{x}' was ignored.[/yellow]",
                True,
            )
        else:
            list_coverages.append(int(x))

    return list_coverages


def percentage_checker(percentages):
    """
    Converts a comma-separated string into a list of floats.
    """
    list_percentages = [float(x.strip()) for x in percentages.split(",")]

    if any(p == 0 for p in list_percentages):
        raise click.UsageError("Percentages must be between > 0.")

    if len(list_percentages) < 4:
        raise click.UsageError("At least four percentages must be provided.")

    # And now add 1 if not persent to the list
    # to calculate CpG number on original sample
    if 1 not in list_percentages:
        list_percentages = list_percentages + [1]

    return list_percentages


def validate_reference(configs):

    if configs.fasta and configs.genome:
        vprint(
            "[yellow]⚠️ Both --fasta and --genome provided. Using the provided fasta file.[/yellow]",
            configs.verbose,
        )

    if configs.fasta:
        if not (configs.fasta.endswith(".fa") or configs.fasta.endswith(".fasta")):
            raise click.UsageError(
                "The fasta file provided must end with .fa or .fasta."
            )
        if not os.path.exists(configs.fasta):
            raise click.UsageError(f"The fasta file '{configs.fasta}' does not exist.")
        return configs.fasta
    else:
        fasta_file = get_reference(configs)
        return fasta_file


def validate_bamdir(bam_dir):
    # Check if directory exists
    if not os.path.exists(bam_dir):
        raise click.UsageError(f"Directory does not exist: {bam_dir}")

    if not os.path.isdir(bam_dir):
        raise click.UsageError(f"Path is not a directory: {bam_dir}")


def ensure_coordinated_sorted(configs):

    # Check if file exists
    if not os.path.exists(configs.bam):
        raise click.UsageError(f"The file '{configs.bam}' does not exist.")

    # Check if file ends with .bam
    if not configs.bam.endswith(".bam"):
        raise click.UsageError("The input file must end with .bam")

    with pysam.AlignmentFile(configs.bam, "rb") as bam:
        sort_order = bam.header.get("HD", {}).get("SO", None)

    if sort_order == "coordinate":
        return configs.bam

    vprint("🔄 BAM file is not coordinate-sorted. Sorting now...", configs.verbose)
    out = configs.bam.replace(".bam", ".csorted.bam")
    cmd = ["samtools", "sort", "-o", out, configs.bam]

    # Run samtools
    subprocess.run(cmd)

    return out


def validate_read_summary(read_summary):
    REQUIRED_COLUMNS = {"Sample", "Percentage", "Read_Count"}
    # Check if file is CSV by extension
    if not read_summary.lower().endswith(".csv"):
        raise click.UsageError(f"The file '{read_summary}' is not a CSV file.")

    # Check file exists
    if not os.path.isfile(read_summary):
        raise click.UsageError(f"File '{read_summary}' does not exist.")

    # Try reading file
    df = pd.read_csv(read_summary, nrows=0)

    # Validate required columns
    missing_cols = REQUIRED_COLUMNS - set(df.columns)
    if missing_cols:
        raise click.UsageError(
            f"Missing required columns in the reads summary file: {', '.join(missing_cols)}"
        )


def validate_cpgs_summary(cpgs_summary):
    REQUIRED_COLUMNS = {"Sample", "Percentage", "Coverage", "CpG_Count"}
    # Check if file is CSV by extension
    if not cpgs_summary.lower().endswith(".csv"):
        raise click.UsageError(f"The file '{cpgs_summary}' is not a CSV file.")

    # Check file exists
    if not os.path.isfile(cpgs_summary):
        raise click.UsageError(f"File '{cpgs_summary}' does not exist.")

    # Try reading file
    df = pd.read_csv(cpgs_summary, nrows=0)

    # Validate required columns
    missing_cols = REQUIRED_COLUMNS - set(df.columns)
    if missing_cols:
        raise click.UsageError(
            f"Missing required columns in the CpGs summary file: {', '.join(missing_cols)}"
        )



import os
import subprocess
import pandas as pd
import pathlib


def test_run_methyldackel(tmp_path, monkeypatch):

    # Init the data and set min coverage to 1
    fasta = pathlib.Path(__file__).parent / "data" / "genome.fa"
    bam_path = pathlib.Path(__file__).parent / "data" / "Ecoli.csorted.bam"
    min_cov = 1

    # Use the BAM filename (without directories) as prefix
    bam_name = os.path.basename(bam_path)
    cov_dir = os.path.join(tmp_path, "covs")
    os.makedirs(cov_dir, exist_ok=True)
    prefix = os.path.join(cov_dir, os.path.splitext(bam_name)[0])
    cmd = ["MethylDackel", "extract", "-o", str(prefix), fasta, bam_path]

    # Run command
    subprocess.run(cmd)

    # Read the file (assuming tab-separated)
    file = prefix + "_CpG.bedGraph"
    df = pd.read_csv(file, sep="\t", header=None, skiprows=1)
    # Count rows where column 5 + column 6 >= 3 (0-based indexing in pandas)
    num_cpgs = int(((df[4] + df[5]) >= min_cov).sum())

    # Check correct stats returned
    assert num_cpgs == 41  # 41 CpGs pass filter



GENOME_URLS = {
    "hg19": "https://hgdownload.soe.ucsc.edu/goldenPath/hg19/bigZips/hg19.fa.gz",
    "hg38": "https://hgdownload.soe.ucsc.edu/goldenPath/hg38/bigZips/hg38.fa.gz",
    "GRCh38": "http://ftp.ensembl.org/pub/release-109/fasta/homo_sapiens/dna/Homo_sapiens.GRCh38.dna.primary_assembly.fa.gz",
    "GRCh37": "http://ftp.ensembl.org/pub/release-109/fasta/homo_sapiens/dna/Homo_sapiens.GRCh37.dna.primary_assembly.fa.gz",
    "mm10": "https://hgdownload.soe.ucsc.edu/goldenPath/mm10/bigZips/mm10.fa.gz",
    "mm39": "https://hgdownload.soe.ucsc.edu/goldenPath/mm10/bigZips/mm39.fa.gz",
}


class ConfigFormatter:

    def __init__(self, **kwargs):
        self.bam = kwargs.get("bam")
        self.bamdir = kwargs.get("bamdir")
        self.outdir = kwargs.get("outdir")
        self.fasta = kwargs.get("fasta")
        self.genome = kwargs.get("genome")
        self.downsampling_percentages = kwargs.get("downsampling_percentages")
        self.minimum_coverage = kwargs.get("minimum_coverage")
        self.threads = kwargs.get("threads")
        self.keep_temporary_files = kwargs.get("keep_temporary_files")
        self.cpgs_file = kwargs.get("cpgs_file")
        self.reads_file = kwargs.get("reads_file")
        self.verbose = kwargs.get("verbose")



import pytest
from unittest.mock import patch, ANY
import gzip
from methurator.config_utils.download_reference import get_reference
from methurator.config_utils.config_formatter import GENOME_URLS
from methurator.config_utils.config_formatter import ConfigFormatter


def test_invalid_fasta_extension(tmp_path):
    wrong = tmp_path / "ref.txt"
    wrong.write_text("dummy")
    configs = ConfigFormatter(
        **{
            "fasta": str(wrong),
            "outdir": str(tmp_path),
            "genome": None,
            "verbose": False,
        }
    )
    with pytest.raises(Exception):
        get_reference(configs)


def test_missing_fasta_file(tmp_path):
    missing = tmp_path / "missing.fa"
    configs = ConfigFormatter(
        **{
            "fasta": str(missing),
            "outdir": str(tmp_path),
            "genome": None,
            "verbose": False,
        }
    )
    with pytest.raises(Exception):
        get_reference(configs)


@patch("methurator.config_utils.download_reference.subprocess.run")
@patch("methurator.config_utils.download_reference.urllib.request.urlretrieve")
def test_download_reference(url_mock, subproc_mock, tmp_path, monkeypatch):
    genome = "hg19"
    genome_url = GENOME_URLS[genome]
    url_mock.return_value = None  # no actual download

    # Fake gz file
    gz_file = tmp_path / f"{genome}.fa.gz"
    with gzip.open(gz_file, "wb") as f:
        f.write(b">seq\nATCG\n")

    # Simulate genome download
    configs = ConfigFormatter(
        **{"fasta": None, "outdir": str(tmp_path), "genome": genome, "verbose": False}
    )
    result = get_reference(configs)

    # Ensure download called correctly
    url_mock.assert_called_once_with(
        genome_url,
        str(gz_file),
        reporthook=ANY,  # ignore progress callback
    )

    # Ensure gunzip executed
    subproc_mock.assert_called_once()
    assert result.endswith(".fa")



import pytest
import rich_click as click
from methurator.config_utils.validation_utils import mincoverage_checker


def test_valid_coverages():
    assert mincoverage_checker("10,5,3") == [10, 5, 3]


def test_values_with_spaces():
    assert mincoverage_checker(" 8 , 4 , 2 ") == [8, 4, 2]


def test_zero_values_are_ignored(monkeypatch):
    # Capture verbose warning output
    messages = []

    def fake_vprint(msg, verbose):
        messages.append(msg)

    monkeypatch.setattr("methurator.config_utils.verbose_utils.vprint", fake_vprint)

    result = mincoverage_checker("5,0,3,0,2")
    assert result == [5, 3, 2]


@pytest.mark.parametrize("bad_value", ["a", "3.5", "-2", "abc"])
def test_invalid_values_raise_exception(bad_value):
    with pytest.raises(click.UsageError):
        mincoverage_checker(f"5,{bad_value},3")


def test_only_zero_values_returns_empty_list(monkeypatch):
    monkeypatch.setattr(
        "methurator.config_utils.verbose_utils.vprint", lambda *args, **kwargs: None
    )
    assert mincoverage_checker("0,0,0") == []



import pytest
import rich_click as click
from methurator.config_utils.validation_utils import percentage_checker


def test_valid_percentages_adds_one():
    result = percentage_checker("0.5, 0.2, 0.2, 0.1")
    assert result == [0.5, 0.2, 0.2, 0.1, 1]


def test_valid_percentages_with_one_present():
    result = percentage_checker("0.7, 0.2, 0.1, 1")
    assert result == [0.7, 0.2, 0.1, 1]  # no extra 1 added


def test_fails_if_any_zero():
    with pytest.raises(click.UsageError):
        percentage_checker("0.5, 0, 0.3, 0.2")


def test_fails_if_less_than_four_values():
    with pytest.raises(click.UsageError):
        percentage_checker("0.5, 0.3, 0.2")


def test_strips_whitespace_correctly():
    result = percentage_checker(" 0.4 , 0.3 ,0.2 ,0.1 ")
    assert result == [0.4, 0.3, 0.2, 0.1, 1]



import pandas as pd
import os
from methurator.plot_utils.plot_checker import plot_checker


def plot_curve(configs):

    # Takes in input the CpGs and Reads stats dataframes and merge them
    cpgs_file = pd.read_csv(configs.cpgs_file)
    reads_file = pd.read_csv(configs.reads_file)
    data = pd.merge(cpgs_file, reads_file, on=["Sample", "Percentage"])

    # Loop over each sample and minimum coverage value to create plots
    for sample in data["Sample"].unique():
        sample_data = data[data["Sample"] == sample]
        for min_val in sample_data["Coverage"].unique():

            # Subset and sort data for plotting
            subset = sample_data[sample_data["Coverage"] == min_val].sort_values(
                by="Percentage"
            )

            # Create output directory for plots if not exists
            plot_dir = os.path.join(configs.outdir, "plots")
            os.makedirs(plot_dir, exist_ok=True)

            # Define plot path
            plot_path = f"{plot_dir}/{sample}_{min_val}x_plot.svg"

            # Checks whether the model fits correctly, then generate plot
            plot_checker(subset, plot_path)



"""
methsaturator
Package for sequencing saturation analysis of
sequencing methylation data.
"""

__version__ = "0.1.0"

from .plot import plot
from .downsample import downsample

__all__ = [
    "plot",
    "downsample",
    "__version__",
]



import rich_click as click
from methurator.plot import plot
from methurator.downsample import downsample


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
def entry_point():
    pass


# Register the 2 subcommands: downsample and plot
entry_point.add_command(plot)
entry_point.add_command(downsample)

if __name__ == "__main__":
    entry_point()



import subprocess
import os
import pandas as pd


def run_methyldackel(bam_path, pct, configs, cpgs_df):

    # Create covs directory to store coverage files
    cov_dir = os.path.join(configs.outdir, "covs")
    os.makedirs(cov_dir, exist_ok=True)

    # Use the BAM filename (without directories) as prefix
    bam_name = os.path.basename(bam_path)
    prefix = os.path.join(cov_dir, os.path.splitext(bam_name)[0])
    cmd = [
        "MethylDackel",
        "extract",
        "-@",
        str(configs.threads),
        "-o",
        str(prefix),
        configs.fasta,
        bam_path,
    ]

    # Run command
    subprocess.run(cmd)

    # Read the file dumped by MethylDackel
    file = prefix + "_CpG.bedGraph"
    df = pd.read_csv(file, sep="\t", header=None, skiprows=1)

    # gzip the file and return the stats
    subprocess.run(["gzip", "-f", f"{prefix}_CpG.bedGraph"])
    sample_name = os.path.basename(bam_path).split(".", 1)[0]
    for min_cov in configs.coverages:
        # Count rows where column 5 + column 6 >= min cov param (0-based indexing in pandas)
        num_cpgs = int(((df[4] + df[5]) >= min_cov).sum())
        cpgs_stats = [sample_name, pct, min_cov, num_cpgs]
        cpgs_df.loc[len(cpgs_df)] = cpgs_stats

    return cpgs_df



from methurator.downsample_utils.methyldackel import run_methyldackel
from methurator.downsample_utils.subsample_bam import subsample_bam
from methurator.config_utils.verbose_utils import vprint
from rich.console import Console
from rich.panel import Panel
import pandas as pd

console = Console()


def run_processing(csorted_bams, configs):

    # Create empty dataframes to store results
    reads_df = pd.DataFrame(columns=["Sample", "Percentage", "Read_Count"])
    cpgs_df = pd.DataFrame(columns=["Sample", "Percentage", "Coverage", "CpG_Count"])

    # Loop over bam files, downsampling percentages and minimum coverage values
    for bam in csorted_bams:

        console.print(
            Panel(
                f"[bold white]🚀 Running saturation analysis on {bam}[/bold white]",
                border_style="green",
            )
        )
        for pct in configs.percentages:
            vprint(
                f"[bold]Subsampling BAM file at[/bold] [cyan]{pct}...[/cyan]",
                configs.verbose,
            )
            results_subsampling_bam, sub_bam = subsample_bam(bam, pct, configs.outdir)
            reads_df.loc[len(reads_df)] = results_subsampling_bam
            vprint("[green]✔ Subsampling completed![/green]", configs.verbose)

            vprint("[bold]Running MethylDackel...[/bold]", configs.verbose)

            cpgs_df = run_methyldackel(sub_bam, pct, configs, cpgs_df)
            vprint("[green]✔ MethylDackel completed![/green]", configs.verbose)

    return cpgs_df, reads_df



import numpy as np

# ===============================================================
# Mathematical Model
# ===============================================================


def asymptotic_growth(x, beta0, beta1):
    """Asymptotic growth model using the arctangent function."""
    return beta0 * np.arctan(beta1 * x)


def derivative_asymptotic_growth(x, beta0, beta1):
    """Derivative of the asymptotic growth model."""
    return beta0 * beta1 / (1 + (beta1 * x) ** 2)


def find_asymptote(params):
    """Return the asymptote value (y-limit as x → ∞)."""
    beta0, _ = params
    return beta0 * np.pi / 2



from methurator.downsample_utils.run_processing import run_processing
from methurator.config_utils.config_formatter import ConfigFormatter
from methurator.config_utils.config_validator import validate_parameters
from methurator.config_utils.bam_dir_utils import bam_to_list
from methurator.config_utils.verbose_utils import vprint
import rich_click as click
from rich.console import Console
from rich.panel import Panel
import os
import shutil


console = Console()


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option(
    "--bam",
    type=click.Path(exists=True),
    required=False,
    help="BAM input file to compute methylation saturation.",
)
@click.option(
    "--bamdir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    required=False,
    help="Directory containing multiple BAM files.",
)
@click.option(
    "--outdir",
    type=click.Path(),
    default="output",
    help="Default ./output directory.",
)
@click.option(
    "--fasta",
    type=click.Path(exists=True),
    help="Fasta file of the reference genome used to align the samples. "
    "If not provided, it will download it according to the specified genome.",
)
@click.option(
    "--genome",
    type=click.Choice(["hg19", "hg38", "GRCh37", "GRCh38", "mm10", "mm39"]),
    default=None,
    help="Genome used to align the samples.",
)
@click.option(
    "--downsampling-percentages",
    "-ds",
    default="0.1,0.25,0.5,0.75",
    help="Percentages used to downsample the .bam file. Default: 0.1,0.25,0.5,0.75",
)
@click.option(
    "--minimum-coverage",
    "-mc",
    default="3",
    help="Minimum CpG coverage to estimate sequencing saturation. It can be either a single integer or a list of integers (e.g 1,3,5). Default: 3",
)
@click.option(
    "--threads",
    "-@",
    type=int,
    default=os.cpu_count() - 2,
    help="Number of threads to use. Default: all available threads - 2.",
)
@click.option(
    "--keep-temporary-files",
    "-k",
    is_flag=True,
    help="If set to True, temporary files will be kept after the analysis. Default: False",
)
@click.option("-v", "--verbose", is_flag=True, help="Enable verbose logging.")
def downsample(**kwargs):

    # Import the parameters and validate them
    configs = ConfigFormatter(**kwargs)
    validate_parameters(configs)

    # Print I/O parameters
    params_text = ""
    params_text += f"[purple]Output directory:[/purple] [blue]{configs.outdir}[/blue]\n"
    if configs.fasta is not None:
        params_text += (
            f"[purple]Reference FASTA:[/purple] [blue]{configs.fasta}[/blue]\n"
        )
    if configs.genome is not None:
        params_text += f"[purple]Genome:[/purple] [blue]{configs.genome}[/blue]\n"
    params_text += f"[purple]Downsampling percentages:[/purple] [blue]{configs.downsampling_percentages}[/blue]\n"
    params_text += f"[purple]Minimum coverage values:[/purple] [blue]{configs.minimum_coverage}[/blue]\n"
    params_text += f"[purple]Threads:[/purple] [blue]{configs.threads}[/blue]\n"
    params_text += f"[purple]Keep temporary files:[/purple] [blue]{configs.keep_temporary_files}[/blue]"
    console.print(
        Panel(
            params_text,
            title="📌 [bold cyan]Input / Output Parameters[/bold cyan]",
            border_style="cyan",
            expand=False,
        )
    )

    # Load bam file(s) and run the downsampling
    csorted_bams = bam_to_list(configs)
    cpgs_df, reads_df = run_processing(csorted_bams, configs)
    reads_df.to_csv(os.path.join(configs.outdir, "reads_summary.csv"), index=False)
    cpgs_df.to_csv(os.path.join(configs.outdir, "cpgs_summary.csv"), index=False)
    vprint(f"[bold] ✅ Dumped summary files to {configs.outdir}.[/bold]", True)

    # Clean-up
    if not configs.keep_temporary_files:
        shutil.rmtree(os.path.join(configs.outdir, "bams"))
        shutil.rmtree(os.path.join(configs.outdir, "covs"))



import rich_click as click
import os
from methurator.config_utils.validation_utils import (
    mincoverage_checker,
    percentage_checker,
    validate_reference,
    validate_bamdir,
)
from methurator.config_utils.verbose_utils import vprint


def validate_parameters(configs):

    # Enforce that at least one of --fasta or --genome is provided
    if configs.fasta is None and configs.genome is None:
        raise click.UsageError(
            "Error: you must provide in input either --fasta or --genome"
        )

    # Enforce that at least one of --bam or --bamdir is provided
    if configs.bam is None and configs.bamdir is None:
        raise click.UsageError(
            "Error: you must provide in input either --bam or --bamdir"
        )

    # Check that downsampling percentages and minimum coverage values are valid
    try:
        configs.percentages = percentage_checker(configs.downsampling_percentages)
    except ValueError as e:
        raise click.UsageError(f"{e}")
    try:
        configs.coverages = mincoverage_checker(configs.minimum_coverage)
    except ValueError as e:
        raise click.UsageError(f"{e}")

    # Run checks on fasta file if provided or download it
    try:
        configs.fasta = validate_reference(configs)
    except ValueError as e:
        raise click.UsageError(f"{e}")

    # Run check on the bam directory if provided
    if configs.bamdir:
        validate_bamdir(configs.bamdir)

    # Create output directory if it doesn't exist
    if not os.path.exists(configs.outdir):
        os.makedirs(configs.outdir, exist_ok=True)
        vprint(
            f"[bold]Created output directory {configs.outdir}...[/bold]",
            configs.verbose,
        )



import pandas as pd
import pytest
import click
from methurator.config_utils.validation_utils import (
    validate_read_summary,
    validate_cpgs_summary,
)


def test_validate_read_summary_valid(tmp_path):
    """Test that a valid CSV with all required columns passes validation."""
    file_path = tmp_path / "reads_summary.csv"
    df = pd.DataFrame(columns=["Sample", "Percentage", "Read_Count"])
    df.to_csv(file_path, index=False)

    # Should not raise
    validate_read_summary(str(file_path))


def test_validate_read_summary_wrong_extension(tmp_path):
    """Test that a non-CSV file raises a UsageError."""
    file_path = tmp_path / "reads_summary.txt"
    file_path.write_text("Sample,Percentage,Read_Count\n")

    with pytest.raises(click.UsageError, match="not a CSV file"):
        validate_read_summary(str(file_path))


def test_validate_read_summary_missing_columns(tmp_path):
    """Test that missing columns raise a UsageError."""
    file_path = tmp_path / "reads_summary.csv"
    df = pd.DataFrame(columns=["Sample", "Read_Count"])  # Missing 'Percentage'
    df.to_csv(file_path, index=False)

    with pytest.raises(click.UsageError, match="Missing required columns"):
        validate_read_summary(str(file_path))


def test_validate_cpgs_summary_valid(tmp_path):
    """Test that a valid CSV with all required columns passes validation."""
    file_path = tmp_path / "cpgs_summary.csv"
    df = pd.DataFrame(columns=["Sample", "Percentage", "Coverage", "CpG_Count"])
    df.to_csv(file_path, index=False)

    # Should not raise
    validate_cpgs_summary(str(file_path))


def test_validate_cpgs_summary_wrong_extension(tmp_path):
    """Test that a non-CSV file raises a UsageError."""
    file_path = tmp_path / "cpgs_summary.tsv"
    file_path.write_text("Sample,Percentage,Coverage,CpG_Count\n")

    with pytest.raises(click.UsageError, match="not a CSV file"):
        validate_cpgs_summary(str(file_path))


def test_validate_cpgs_summary_missing_columns(tmp_path):
    """Test that missing columns raise a UsageError."""
    file_path = tmp_path / "cpgs_summary.csv"
    df = pd.DataFrame(
        columns=["Sample", "Percentage", "CpG_Count"]
    )  # Missing 'Coverage'
    df.to_csv(file_path, index=False)

    with pytest.raises(click.UsageError, match="Missing required columns"):
        validate_cpgs_summary(str(file_path))



import subprocess
import os


def subsample_bam(bam_path, percentage, output_dir, seed=42):
    """
    Subsamples a BAM file using samtools according to a percentage.
    """

    # Create bams directory to store the bam files
    bam_dir = os.path.join(output_dir, "bams")
    os.makedirs(bam_dir, exist_ok=True)

    # Round the downsampling percentage and init the output bam name
    round_pct = round(percentage, 2)
    output_path = bam_path

    # If percentage is different than 1, then subsample
    if percentage != 1:

        # Extract the sample name and create the output path
        base_name = os.path.basename(bam_path).replace(
            ".bam", f"_subsample_{round_pct:.1f}.bam"
        )
        output_path = os.path.join(bam_dir, base_name)

        # Subsample BAM using samtools
        cmd = [
            "samtools",
            "view",
            "-s",
            str(round_pct),
            "--subsample-seed",
            str(seed),
            "-b",
            bam_path,
            "-o",
            output_path,
        ]
        subprocess.run(cmd, check=True)

    # Index the subsampled BAM
    cmd2 = ["samtools", "index", output_path]
    subprocess.run(cmd2, check=True)

    # Count reads in the subsampled BAM
    read_count = int(
        subprocess.run(
            ["samtools", "view", "-c", output_path], capture_output=True, text=True
        ).stdout.strip()
    )

    # Extract the name and save it in the stats list
    sample_name = os.path.basename(bam_path).split(".", 1)[0]
    sample_stats = [sample_name, round_pct, read_count]
    return sample_stats, output_path



import os
import urllib.request
import subprocess
from methurator.config_utils.config_formatter import GENOME_URLS
from tqdm import tqdm
from methurator.config_utils.verbose_utils import vprint


class DownloadProgressBar(tqdm):
    def update_to(self, blocks=1, block_size=1, total_size=None):
        if total_size is not None:
            self.total = total_size
        self.update(blocks * block_size - self.n)


def get_reference(configs):

    # Download the reference genome
    url = GENOME_URLS[configs.genome]
    gz_path = os.path.join(configs.outdir, f"{configs.genome}.fa.gz")
    fasta_path = os.path.join(configs.outdir, f"{configs.genome}.fa")

    vprint(
        f"[bold]Downloading reference genome[/bold] [green]'{configs.genome}'...[/green]",
        True,
    )
    with DownloadProgressBar(
        unit="B", unit_scale=True, miniters=1, desc="Downloading"
    ) as t:
        urllib.request.urlretrieve(url, gz_path, reporthook=t.update_to)
    vprint(f"🗜️  Extracting {gz_path}...", configs.verbose)
    subprocess.run(["gunzip", "-f", gz_path], check=True)
    vprint(f"✅ Reference genome ready: {fasta_path}", True)

    return fasta_path



import numpy as np
from rich.console import Console
from methurator.plot_utils.math_model import asymptotic_growth
import plotly.graph_objs as go

# ===============================================================
# Plotting Functions
# ===============================================================

console = Console()


def human_readable(num):
    """Format large numbers into human-readable form (e.g., 1.5K, 2.3M, 1.1B)."""
    for unit, divisor in [("B", 1e9), ("M", 1e6), ("K", 1e3)]:
        if num >= divisor:
            return f"{num/divisor:.1f}{unit}"
    return str(int(num))


def fmt_list(values):
    """Round and convert a list of numbers to human-readable strings."""
    return [human_readable(round(v, 0)) for v in values]


def make_base_plot(title, xaxis="Number of reads", yaxis="Number of CpGs"):
    """Create a preformatted Plotly figure."""
    fig = go.Figure()
    fig.update_xaxes(title_text=xaxis, showgrid=True)
    fig.update_yaxes(title_text=yaxis, showgrid=True)
    fig.update_layout(
        title=dict(text=title, x=0.5, y=0.9),
        width=950,
        height=620,
        template="plotly_white",
        legend=dict(yanchor="bottom", y=0.01, xanchor="right", x=0.99),
    )
    return fig


def plot_fallback(plot_obj):
    """Fallback plot when curve fitting fails."""
    x_reads = fmt_list([x * plot_obj.reads for x in plot_obj.x_data])
    y_cpgs = fmt_list(plot_obj.y_data)

    fig = make_base_plot(f"{plot_obj.title}<br><sup>{plot_obj.error_msg}</sup>")
    fig.add_trace(
        go.Scatter(
            x=x_reads,
            y=plot_obj.y_data,
            mode="lines",
            name="Observed data",
            customdata=np.column_stack((y_cpgs, plot_obj.x_data)),
            hovertemplate=(
                "Number of reads: %{x}<br>"
                "Number CpGs: %{customdata[0]}<br>"
                "Downsampling: %{customdata[1]}"
                "<extra></extra>"
            ),
        )
    )

    html_path = plot_obj.output_path.replace(".svg", ".html")
    fig.write_html(html_path)
    print(f"⚠️ Curve fitting failed but plot saved anyway to: {html_path}.")


def plot_fitted_data(plot_obj):
    """Plot observed and predicted CpG values based on fitted asymptotic growth."""
    # Extended x-data for predictions
    x_all = np.append(plot_obj.x_data, np.linspace(1.2, 2.0, 5))
    y_all = asymptotic_growth(x_all, *plot_obj.params)

    reads_all = [x * plot_obj.reads for x in x_all]
    reads_fmt = fmt_list(reads_all)
    cpgs_fmt = fmt_list(y_all)
    saturation = np.round((y_all / plot_obj.asymptote) * 100, 0)

    n_obs = len(plot_obj.x_data) - 1

    obs = slice(None, n_obs + 1)
    pred = slice(n_obs, None)
    fig = make_base_plot(plot_obj.title)

    # Observed
    fig.add_trace(
        go.Scatter(
            x=reads_fmt[obs],
            y=y_all[obs],
            mode="lines",
            name="Observed data",
            customdata=np.column_stack((cpgs_fmt[obs], saturation[obs], x_all[obs])),
            hovertemplate=(
                "Number of reads: %{x}<br>"
                "Number CpGs: %{customdata[0]}<br>"
                "Saturation: %{customdata[1]}%<br>"
                "Downsampling: %{customdata[2]}"
                "<extra></extra>"
            ),
        )
    )

    # Predicted
    fig.add_trace(
        go.Scatter(
            x=reads_fmt[pred],
            y=y_all[pred],
            mode="lines+markers",
            name="Predicted data",
            customdata=np.column_stack((cpgs_fmt[pred], saturation[pred])),
            hovertemplate=(
                "Number of reads: %{x}<br>"
                "Number CpGs: %{customdata[0]}<br>"
                "Saturation: %{customdata[1]}%"
                "<extra></extra>"
            ),
            line=dict(dash="dash"),
        )
    )

    # Asymptote line
    fig.add_hline(
        y=plot_obj.asymptote,
        line_dash="dash",
        annotation_text=f"Asymptote = {human_readable(plot_obj.asymptote)} CpGs",
        annotation_position="bottom right",
    )

    html_path = plot_obj.output_path.replace(".svg", ".html")
    fig.write_html(html_path)
    print(f"✅ Plot saved to: {html_path}")



import pytest
from methurator.config_utils.config_formatter import ConfigFormatter
from methurator.config_utils.validation_utils import ensure_coordinated_sorted
import pathlib


@pytest.fixture
def bam_file(tmp_path):
    bam_path = pathlib.Path(__file__).parent / "data" / "Ecoli.unsorted.bam"
    return str(bam_path)


def test_file_not_found():
    configs = ConfigFormatter(**{"bam": "does_not_exist.bam", "verbose": False})
    with pytest.raises(Exception):
        ensure_coordinated_sorted(configs)


def test_wrong_extension(bam_file):
    wrong = bam_file.replace(".bam", ".txt")
    configs = ConfigFormatter(**{"bam": wrong, "verbose": False})
    with pytest.raises(Exception):
        ensure_coordinated_sorted(configs)


def test_unsorted_bam_triggers_sort(monkeypatch, bam_file):
    calls = []
    configs = ConfigFormatter(**{"bam": bam_file, "verbose": False})

    def fake_run(*args, **kwargs):
        calls.append(args[0])

    monkeypatch.setattr("subprocess.run", fake_run)

    result = ensure_coordinated_sorted(configs)

    assert calls, "samtools not called for unsorted BAM"
    assert result.endswith(".csorted.bam")



import warnings
import numpy as np
from scipy.optimize import curve_fit, OptimizeWarning
from methurator.plot_utils.math_model import asymptotic_growth, find_asymptote
from methurator.plot_utils.plot_functions import plot_fitted_data, plot_fallback


class PlotObject:
    def __init__(self, output_path):
        self.x_data = []
        self.y_data = []
        self.asymptote = str
        self.params = []
        self.title = str
        self.reads = int
        self.error_msg = None
        self.output_path = output_path


def plot_checker(data, output_path):

    # Create the PlotObject
    plot_obj = PlotObject(output_path)

    # Add the zeros at the beginning of the data to fit the model
    plot_obj.x_data = np.array([0] + data["Percentage"].tolist())
    plot_obj.y_data = np.array([0] + data["CpG_Count"].tolist())

    # try to fit the asymptotic growth model to the data
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("error", OptimizeWarning)
            plot_obj.params, _ = curve_fit(
                asymptotic_growth, plot_obj.x_data, plot_obj.y_data, p0=[1, 1]
            )
            plot_obj.asymptote = find_asymptote(plot_obj.params)
            fit_success = True
    # If not enough data points or fitting fails, handle the exception
    except (RuntimeError, OptimizeWarning) as e:
        fit_success = False
        plot_obj.params, plot_obj.asymptote, plot_obj.error_msg = None, None, str(e)

    # Prepare title and reads information
    plot_obj.title = data["Sample"].iloc[0]
    plot_obj.reads = int(data["Read_Count"].iloc[-1])

    # If fitting was successful, plot the fitted curve;
    # otherwise, plot the dots with the error message
    if fit_success:
        plot_fitted_data(plot_obj)
    else:
        plot_fallback(plot_obj)



import os
import pathlib
from methurator.downsample_utils.subsample_bam import subsample_bam


def test_subsample_when_percentage_is_one(tmp_path, monkeypatch):

    # Specify percentage = 1 and the bam file path
    pct = 1
    bam = pathlib.Path(__file__).parent / "data" / "Ecoli.csorted.bam"

    # Extract the sample name and create the output path
    base_name = os.path.basename(bam).replace(".bam", f"_subsample_{pct}.bam")
    output_path = os.path.join(tmp_path, "bams/", base_name)

    # Run the subsampling and get the output
    stats, output = subsample_bam(str(bam), pct, str(output_path))

    # Check that the output are the expected ones
    assert stats[0] == "Ecoli"
    assert stats[1] == 1
    assert stats[2] == 10
    assert output == str(bam)


def test_subsample_when_percentage_not_one(tmp_path, monkeypatch):
    # Specify percentage != 1 and the bam file path
    pct = 0.5
    bam = pathlib.Path(__file__).parent / "data" / "Ecoli.csorted.bam"

    # Extract the sample name and create the output path
    base_name = os.path.basename(bam).replace(".bam", f"_subsample_{pct}.bam")
    output_path = os.path.join(tmp_path, "bams/", base_name)

    # Run the subsampling and get the output
    stats, output = subsample_bam(str(bam), pct, str(tmp_path), 42)

    # Check that the output are the expected ones
    assert stats[0] == "Ecoli"
    assert stats[1] == 0.5
    assert stats[2] == 8  # total number of reads in the subsampled bam
    assert output == str(output_path)
